/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.mapstruct.Mapper;
import org.mapstruct.Mapping;
import org.mapstruct.MappingConstants;
import org.mapstruct.ReportingPolicy;
import org.mapstruct.ValueMapping;
import org.springframework.beans.factory.annotation.Autowired;

import de.ozgcloud.nachrichten.postfach.PostfachNachricht.Direction;
import de.ozgcloud.vorgang.common.GrpcObject;
import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItem;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Mapper(unmappedTargetPolicy = ReportingPolicy.WARN)
public abstract class PostfachNachrichtMapper {

	@Autowired
	private GrpcObjectMapper grpcObjectMapper;

	@Mapping(target = "attachments", source = "attachmentList")
	@Mapping(target = "createdAt", ignore = true)
	@Mapping(target = "createdBy", ignore = true)
	@Mapping(target = "direction", ignore = true)
	@Mapping(target = "messageId", ignore = true)
	@Mapping(target = "sentAt", ignore = true)
	@Mapping(target = "sentSuccessful", ignore = true)
	@ValueMapping(source = "UNRECOGNIZED", target = MappingConstants.NULL)
	@ValueMapping(source = "UNDEFINED", target = MappingConstants.NULL)
	public abstract PostfachNachricht fromGrpcMail(GrpcPostfachMail grpcMail);

	public PostfachAddressIdentifier toPostfachAddressIdentifier(GrpcObject grpcObject) {
		var postfachId = grpcObject.getPropertyList().get(0).getValue(0);
		return StringBasedIdentifier.builder().postfachId(postfachId).build();
	}

	public GrpcPostfachMail fromMap(Map<String, Object> mailMap) {
		var postfachMailBuilder = GrpcPostfachMail.newBuilder()
				.setId(MapUtils.getString(mailMap, PostfachNachricht.FIELD_ID))
				.setVorgangId(MapUtils.getString(mailMap, PostfachNachricht.FIELD_VORGANG_ID))
				.setReferencedNachricht(MapUtils.getString(mailMap, PostfachNachricht.REFERENCED_NACHRICHT_FIELD, StringUtils.EMPTY))
				.setPostfachId(MapUtils.getString(mailMap, PostfachNachricht.FIELD_POSTFACH_ID, StringUtils.EMPTY))
				.setCreatedAt(MapUtils.getString(mailMap, PostfachNachricht.FIELD_CREATED_AT))
				.setCreatedBy(MapUtils.getString(mailMap, PostfachNachricht.FIELD_CREATED_BY, StringUtils.EMPTY))
				.setSentAt(MapUtils.getString(mailMap, PostfachNachricht.FIELD_SENT_AT, StringUtils.EMPTY))
				.setSentSuccessful(MapUtils.getBooleanValue(mailMap, PostfachNachricht.FIELD_SENT_SUCCESSFUL, false))
				.setMessageCode(MapUtils.getString(mailMap, PostfachNachricht.FIELD_MESSAGE_CODE, StringUtils.EMPTY))
				.setDirection(GrpcDirection.valueOf(MapUtils.getString(mailMap, PostfachNachricht.FIELD_DIRECTION)))
				.setSubject(MapUtils.getString(mailMap, PostfachNachricht.FIELD_SUBJECT))
				.setMailBody(MapUtils.getString(mailMap, PostfachNachricht.FIELD_MAIL_BODY))
				.setReplyOption(MapUtils.getString(mailMap, PostfachNachricht.FIELD_REPLY_OPTION));

		Optional.ofNullable(mailMap.get(PostfachNachricht.FIELD_ATTACHMENTS))
				.filter(this::doesAttachmentsExists)
				.map(this::getAttachments)
				.ifPresent(postfachMailBuilder::addAllAttachment);

		Optional.ofNullable(getAsMap(mailMap, PostfachNachricht.POSTFACH_ADDRESS_FIELD))
				.filter(MapUtils::isNotEmpty)
				.map(this::buildGrpcPostfachAddress)
				.ifPresent(postfachMailBuilder::setPostfachAddress);

		return postfachMailBuilder.build();
	}

	private boolean doesAttachmentsExists(Object attachments) {
		return (attachments instanceof String attachment && StringUtils.isNotBlank(attachment)) || attachments instanceof Collection;
	}

	private Iterable<String> getAttachments(Object attachments) {
		if (attachments instanceof String attachment) {
			return Collections.singletonList(attachment);
		}
		return (Iterable<String>) attachments;
	}

	GrpcPostfachAddress buildGrpcPostfachAddress(Map<String, Object> postfachAddressMap) {
		var postfachAddressBuilder = GrpcPostfachAddress.newBuilder()
				.setType(MapUtils.getIntValue(postfachAddressMap, PostfachAddress.TYPE_FIELD))
				.setVersion(MapUtils.getString(postfachAddressMap, PostfachAddress.VERSION_FIELD))
				.setIdentifier(grpcObjectMapper.fromMap(postfachAddressMap));
		Optional.ofNullable(MapUtils.getString(postfachAddressMap, PostfachAddress.SERVICEKONTO_TYPE_FIELD))
				.ifPresent(postfachAddressBuilder::setServiceKontoType);
		return postfachAddressBuilder.build();
	}

	public PostfachNachricht fromAttachedItem(GrpcVorgangAttachedItem grpcVorgangAttachedItem) {
		return fromMapToPostfachMail(grpcObjectMapper.mapFromGrpc(grpcVorgangAttachedItem.getItem()));
	}

	PostfachNachricht fromMapToPostfachMail(Map<String, Object> mailMap) {
		var postfachMailBuilder = PostfachNachricht.builder()
				.id(MapUtils.getString(mailMap, PostfachNachricht.FIELD_ID))
				.vorgangId(MapUtils.getString(mailMap, (PostfachNachricht.FIELD_VORGANG_ID)))
				.referencedNachricht(MapUtils.getString(mailMap, PostfachNachricht.REFERENCED_NACHRICHT_FIELD, StringUtils.EMPTY))
				.postfachId(MapUtils.getString(mailMap, PostfachNachricht.FIELD_POSTFACH_ID, StringUtils.EMPTY))
				.messageId(MapUtils.getString(mailMap, PostfachNachricht.FIELD_MESSAGE_ID))
				.createdAt(ZonedDateTime.parse(MapUtils.getString(mailMap, PostfachNachricht.FIELD_CREATED_AT)))
				.createdBy(MapUtils.getString(mailMap, PostfachNachricht.FIELD_CREATED_BY))
				.sentAt(getString(mailMap, PostfachNachricht.FIELD_SENT_AT).map(ZonedDateTime::parse).orElse(null))
				.sentSuccessful(MapUtils.getBoolean(mailMap, PostfachNachricht.FIELD_SENT_SUCCESSFUL))
				.messageCode(MapUtils.getString(mailMap, PostfachNachricht.FIELD_MESSAGE_CODE))
				.direction(Direction.valueOf(MapUtils.getString(mailMap, PostfachNachricht.FIELD_DIRECTION)))
				.subject(MapUtils.getString(mailMap, PostfachNachricht.FIELD_SUBJECT))
				.mailBody(MapUtils.getString(mailMap, PostfachNachricht.FIELD_MAIL_BODY))
				.replyOption(PostfachNachricht.ReplyOption.valueOf(MapUtils.getString(mailMap, PostfachNachricht.FIELD_REPLY_OPTION)))
				.attachments(getAsList(mailMap, PostfachNachricht.FIELD_ATTACHMENTS));

		Optional.ofNullable(getAsMap(mailMap, PostfachNachricht.POSTFACH_ADDRESS_FIELD))
				.filter(MapUtils::isNotEmpty)
				.map(this::buildPostfachAddress)
				.ifPresent(postfachMailBuilder::postfachAddress);

		return postfachMailBuilder.build();
	}

	private Optional<String> getString(Map<String, Object> mailMap, String key) {
		return Optional.ofNullable(MapUtils.getString(mailMap, key));
	}

	@SuppressWarnings("unchecked")
	private List<String> getAsList(Map<String, Object> mailMap, String fieldName) {
		var valueObject = mailMap.get(fieldName);
		if (valueObject instanceof Collection<?> values) {
			return (List<String>) values;
		}
		if (valueObject instanceof String value) {
			return List.of(value);
		}
		return Collections.emptyList();
	}

	PostfachAddress buildPostfachAddress(Map<String, Object> postfachAddressMap) {
		return PostfachAddress.builder()
				.type(MapUtils.getIntValue(postfachAddressMap, PostfachAddress.TYPE_FIELD))
				.version(MapUtils.getString(postfachAddressMap, PostfachAddress.VERSION_FIELD))
				.identifier(buildIdentifier(postfachAddressMap))
				.serviceKontoType(MapUtils.getString(postfachAddressMap, PostfachAddress.SERVICEKONTO_TYPE_FIELD))
				.build();
	}

	PostfachAddressIdentifier buildIdentifier(Map<String, Object> postfachAddressMap) {
		return StringBasedIdentifier.builder().postfachId(getPostfachId(postfachAddressMap)).build();
	}

	private String getPostfachId(Map<String, Object> postfachAddressMap) {
		var identifier = getAsMap(postfachAddressMap, PostfachAddress.IDENTIFIER_FIELD);
		return MapUtils.getString(identifier, PostfachNachricht.FIELD_POSTFACH_ID);

	}

	@SuppressWarnings("unchecked")
	Map<String, Object> getAsMap(Map<String, Object> map, String fieldName) {
		return MapUtils.getMap(map, fieldName, Collections.emptyMap());
	}

	public Map<String, Object> mapToMap(PostfachNachricht nachricht) {
		var result = new HashMap<>(Map.of(
				PostfachNachricht.FIELD_VORGANG_ID, nachricht.getVorgangId(),
				PostfachNachricht.FIELD_CREATED_AT, nachricht.getCreatedAt().format(DateTimeFormatter.ISO_DATE_TIME),
				PostfachNachricht.FIELD_DIRECTION, nachricht.getDirection().name(),
				PostfachNachricht.FIELD_REPLY_OPTION, nachricht.getReplyOption().name(),
				PostfachNachricht.FIELD_SUBJECT, nachricht.getSubject(),
				PostfachNachricht.FIELD_MAIL_BODY, nachricht.getMailBody(),
				PostfachNachricht.FIELD_ATTACHMENTS, nachricht.getAttachments()));

		putIfNonNull(result, PostfachNachricht.FIELD_POSTFACH_ID, nachricht.getPostfachId());
		putIfNonNull(result, PostfachNachricht.FIELD_MESSAGE_ID, nachricht.getMessageId());
		putIfNonNull(result, PostfachNachricht.FIELD_CREATED_BY, nachricht.getCreatedBy());
		putIfNonNull(result, PostfachNachricht.FIELD_SENT_SUCCESSFUL, nachricht.getSentSuccessful());
		putIfNonNull(result, PostfachNachricht.FIELD_MESSAGE_CODE, nachricht.getMessageCode());
		putIfNonNull(result, PostfachNachricht.REFERENCED_NACHRICHT_FIELD, nachricht.getReferencedNachricht());
		putDateIfNonNull(result, PostfachNachricht.FIELD_SENT_AT, nachricht.getSentAt());

		Optional.ofNullable(nachricht.getPostfachAddress()).map(this::buildPostfachAddressMap)
				.ifPresent(postfachMaiMap -> result.put(PostfachNachricht.POSTFACH_ADDRESS_FIELD, postfachMaiMap));

		return Collections.unmodifiableMap(result);
	}

	private Map<String, Object> putIfNonNull(Map<String, Object> mapIn, String key, Object value) {
		if (Objects.nonNull(value)) {
			mapIn.put(key, value);
		}
		return mapIn;
	}

	private Map<String, Object> putDateIfNonNull(Map<String, Object> mapIn, String key, Object value) {
		if (Objects.nonNull(value)) {
			mapIn.put(key, ((ZonedDateTime) value).format(DateTimeFormatter.ISO_DATE_TIME));
		}
		return mapIn;
	}

	private Map<String, Object> buildPostfachAddressMap(PostfachAddress postfachAddress) {
		var resultMap = new HashMap<String, Object>();
		resultMap.put(PostfachAddress.IDENTIFIER_FIELD, buildPostfachAddressIdentifierMap(postfachAddress));
		resultMap.put(PostfachAddress.VERSION_FIELD, postfachAddress.getVersion());
		resultMap.put(PostfachAddress.TYPE_FIELD, postfachAddress.getType());
		Optional.ofNullable(postfachAddress.getServiceKontoType()).ifPresentOrElse(
				serviceKontoType -> resultMap.put(PostfachAddress.SERVICEKONTO_TYPE_FIELD, serviceKontoType),
				() -> LOG.warn("ServiceKontoType is null"));
		return Collections.unmodifiableMap(resultMap);
	}

	private Map<String, Object> buildPostfachAddressIdentifierMap(PostfachAddress postfachAddress) {
		var identifier = (StringBasedIdentifier) postfachAddress.getIdentifier();
		return Map.of(PostfachNachricht.FIELD_POSTFACH_ID, identifier.getPostfachId());
	}
}