/*
 * Copyright (c) 2024.
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.nachrichten.antragraum;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import io.grpc.stub.StreamObserver;

class AntragraumGrpcServiceTest {
	@Spy
	@InjectMocks
	private AntragraumGrpcService grpcService;
	@Mock
	private AntragraumService service;
	@Mock
	private AntragraumNachrichtMapper nachrichtMapper;

	@Nested
	class TestFindRueckfragen {

		@Mock
		private StreamObserver<GrpcFindRueckfragenResponse> streamObserver;

		@BeforeEach
		void setup() {
			when(service.findRueckfragen(any())).thenReturn(Stream.of(RueckfrageTestFactory.create()));
			when(nachrichtMapper.toGrpc(any(Rueckfrage.class))).thenReturn(GrpcRueckfrageTestFactory.create());
		}

		@Test
		void shouldCallMapper() {
			grpcService.findRueckfragen(GrpcFindRueckfrageRequestTestFactory.create(), streamObserver);

			verify(nachrichtMapper).toGrpc(any(Rueckfrage.class));
		}

		@Test
		void shouldCallOnNext() {
			grpcService.findRueckfragen(GrpcFindRueckfrageRequestTestFactory.create(), streamObserver);

			verify(streamObserver).onNext(any(GrpcFindRueckfragenResponse.class));
		}

		@Test
		void shouldCallOnCompleted() {
			grpcService.findRueckfragen(GrpcFindRueckfrageRequestTestFactory.create(), streamObserver);

			verify(streamObserver).onCompleted();
		}

		@Test
		void shouldLoadAnswers() {
			grpcService.findRueckfragen(GrpcFindRueckfrageRequestTestFactory.create(), streamObserver);

			verify(service).findAnswers(GrpcFindRueckfrageRequestTestFactory.SAML_TOKEN, GrpcRueckfrageTestFactory.ID);
		}
	}

	@DisplayName("Send rueckfrage answer")
	@Nested
	class TestSendRueckfrageAnswer {
		@Mock
		private StreamObserver<GrpcSendRueckfrageAnswerResponse> streamObserver;
		private PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();
		private Rueckfrage rueckfrage = RueckfrageTestFactory.create();

		@BeforeEach
		void setup() {
			when(nachrichtMapper.fromRueckfrageAnswer(any(GrpcRueckfrageAnswer.class))).thenReturn(postfachNachricht);
			when(service.sendRueckfrageAnswer(anyString(), anyString(), any(PostfachNachricht.class)))
					.thenReturn(UUID.randomUUID().toString());
			doReturn(postfachNachricht).when(grpcService).enrichRueckfrageAnswer(any(), any());
			doReturn(rueckfrage).when(grpcService).getRueckfrage(any(), any());
		}

		@Test
		void shouldCallMapper() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);

			verify(nachrichtMapper).fromRueckfrageAnswer(any(GrpcRueckfrageAnswer.class));
		}

		@Test
		void shouldGetRueckfrage() {
			sendRueckfrageAnswer();

			verify(grpcService).getRueckfrage(GrpcSendRueckfrageAnswerRequestTestFactory.SAML_TOKEN,
					GrpcRueckfrageAnswerTestFactory.RUECKFRAGE_ID);
		}

		@Test
		void shouldEnrichPostfachNachricht() {
			sendRueckfrageAnswer();

			verify(grpcService).enrichRueckfrageAnswer(postfachNachricht, rueckfrage);
		}

		@Test
		void shouldCallAntragraumService() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);

			verify(service).sendRueckfrageAnswer(anyString(), anyString(), any(PostfachNachricht.class));
		}

		@Test
		void shouldCallOnNext() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);

			verify(streamObserver).onNext(any(GrpcSendRueckfrageAnswerResponse.class));
		}

		@Test
		void shouldCallOnCompleted() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);

			verify(streamObserver).onCompleted();
		}

		private void sendRueckfrageAnswer() {
			grpcService.sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequestTestFactory.create(), streamObserver);
		}
	}

	@DisplayName("Get rueckfrage")
	@Nested
	class TestGetRueckfrage {

		private final Rueckfrage matchingRueckfrage = RueckfrageTestFactory.createBuilder().id(GrpcRueckfrageAnswerTestFactory.RUECKFRAGE_ID)
				.build();
		private final Rueckfrage rueckfrage = RueckfrageTestFactory.create();

		@BeforeEach
		void mock() {
			when(service.findRueckfragen(any())).thenReturn(Stream.of(rueckfrage, matchingRueckfrage));
		}

		@Test
		void shouldCallService() {
			getRueckfrage();

			verify(service).findRueckfragen(GrpcSendRueckfrageAnswerRequestTestFactory.SAML_TOKEN);
		}

		@Test
		void shouldReturnMatchingPostfachNachricht() {
			var rueckfrage = getRueckfrage();

			assertThat(rueckfrage).isEqualTo(matchingRueckfrage);
		}

		@Test
		void shouldThrowExceptionIfNoRueckfrageExists() {
			when(service.findRueckfragen(any())).thenReturn(Stream.empty());

			assertThatThrownBy(() -> getRueckfrage()).isInstanceOf(TechnicalException.class);
		}

		private Rueckfrage getRueckfrage() {
			return grpcService.getRueckfrage(GrpcSendRueckfrageAnswerRequestTestFactory.SAML_TOKEN,
					GrpcRueckfrageAnswerTestFactory.RUECKFRAGE_ID);
		}
	}

	@DisplayName("Enrich postfachNachricht")
	@Nested
	class TestEnrichPostfachNachricht {

		private PostfachNachricht postfachNachrichtToEnrich = PostfachNachrichtTestFactory.createBuilder()
				.postfachAddress(null)
				.vorgangId(null)
				.sentAt(null)
				.build();

		private final Rueckfrage postfachNachrichtRueckfrage = RueckfrageTestFactory.create();

		@Test
		void shouldSetVorgangId() {
			var enrichedPostfachNachricht = grpcService.enrichRueckfrageAnswer(postfachNachrichtToEnrich, postfachNachrichtRueckfrage);

			assertThat(enrichedPostfachNachricht.getVorgangId()).isEqualTo(postfachNachrichtRueckfrage.getVorgangId());
		}

		@Test
		void shouldSetPostfachAddresses() {
			var enrichedPostfachNachricht = grpcService.enrichRueckfrageAnswer(postfachNachrichtToEnrich, postfachNachrichtRueckfrage);

			assertThat(enrichedPostfachNachricht.getPostfachAddress()).isEqualTo(postfachNachrichtRueckfrage.getPostfachAddress());
		}

		@Test
		void shouldSetSentAt() {
			var enrichedPostfachNachricht = grpcService.enrichRueckfrageAnswer(postfachNachrichtToEnrich, postfachNachrichtRueckfrage);

			assertThat(enrichedPostfachNachricht.getSentAt()).isCloseTo(ZonedDateTime.now(), within(2, ChronoUnit.SECONDS));
		}
	}
}