package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandId;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.nachrichten.common.vorgang.VorgangService;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;

class PersistPostfachNachrichtServiceImplTest {

	@Spy
	@InjectMocks
	private PersistPostfachNachrichtServiceImpl service;
	@Mock
	private VorgangService vorgangService;
	@Mock
	private OzgCloudCommandService commandService;
	@Mock
	private AttachedItemRemoteService attachedItemRemoteService;
	@Mock
	private PostfachNachrichtMapper postfachNachrichtMapper;

	@DisplayName("Persist answer")
	@Nested
	class TestPersistAnswer {

		@Captor
		private ArgumentCaptor<PostfachNachricht> postfachNachrichtCaptor;

		private final OzgCloudCommand command = OzgCloudCommand.builder().id(OzgCloudCommandId.from(UUID.randomUUID().toString())).build();
		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.createBuilder().referencedNachricht(null).build();

		@BeforeEach
		void mock() {
			doReturn(command).when(service).buildCommand(any());
			when(commandService.create(any())).thenReturn(command);
		}

		@Test
		void shouldCallBuildCommand() {
			service.persistAnswer(GrpcPostfachMailTestFactory.REFERENCED_NACHRICHT_ID, postfachNachricht);

			verify(service).buildCommand(postfachNachrichtCaptor.capture());
			assertThat(postfachNachrichtCaptor.getValue().getReferencedNachricht()).isEqualTo(GrpcPostfachMailTestFactory.REFERENCED_NACHRICHT_ID);
		}

		@Test
		void shouldCallCommandService() {
			service.persistAnswer(GrpcPostfachMailTestFactory.REFERENCED_NACHRICHT_ID, postfachNachricht);

			verify(commandService).create(command);
		}
	}

	@DisplayName("Build command")
	@Nested
	class TestBuildCommand {

		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void mock() {
			doReturn(Map.of("dummyKey", "dummyValue")).when(service).buildCommandBody(any());
		}

		@Test
		void shouldHaveSetOrder() {
			var command = service.buildCommand(postfachNachricht);

			assertThat(command.getOrder()).isEqualTo(PersistPostfachNachrichtServiceImpl.CREATE_ATTACHED_ITEM_ORDER);
		}

		@Test
		void shouldHaveSetVorgangId() {
			var command = service.buildCommand(postfachNachricht);

			assertThat(command.getVorgangId()).hasToString(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldHaveSetRelationId() {
			var command = service.buildCommand(postfachNachricht);

			assertThat(command.getRelationId()).hasToString(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallBuildCommandBody() {
			service.buildCommand(postfachNachricht);

			verify(service).buildCommandBody(postfachNachricht);
		}

		@Test
		void shouldHaveSetBodyObjet() {
			var command = service.buildCommand(postfachNachricht);

			assertThat(command.getBodyObject()).isNotEmpty();
		}
	}

	@DisplayName("Build command body")
	@Nested
	class TestBuildCommandBody {

		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();
		private Map<String, Object> postfachNachrichtAsMap = Map.of("dummyKey", "dummyValue");

		@BeforeEach
		void mock() {
			when(postfachNachrichtMapper.mapToMap(any())).thenReturn(postfachNachrichtAsMap);
		}

		@Test
		void shouldHaveSetClient() {
			var body = service.buildCommandBody(postfachNachricht);

			assertThat(body).containsEntry(PersistPostfachNachrichtServiceImpl.CLIENT_FIELD, PersistPostfachNachrichtServiceImpl.CLIENT);
		}

		@Test
		void shouldHaveSetVorgangId() {
			var body = service.buildCommandBody(postfachNachricht);

			assertThat(body).containsEntry(PersistPostfachNachrichtServiceImpl.VORGANG_ID_FIELD, MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldHaveSetItemName() {
			var body = service.buildCommandBody(postfachNachricht);

			assertThat(body).containsEntry(PersistPostfachNachrichtServiceImpl.ITEM_NAME_FIELD, PersistPostfachNachrichtServiceImpl.ITEM_NAME);
		}

		@Test
		void shouldCallMapper() {
			service.buildCommandBody(postfachNachricht);

			verify(postfachNachrichtMapper).mapToMap(postfachNachricht);
		}

		@Test
		void shouldHaveSetItem() {
			var body = service.buildCommandBody(postfachNachricht);

			assertThat(body).containsEntry(PersistPostfachNachrichtServiceImpl.ITEM_FIELD, postfachNachrichtAsMap);
		}
	}

	@DisplayName("Find answers")
	@Nested
	class TestFindAnswers {

		private final String serviceKontoType = "serviceKontoTypeDummy";

		private final PostfachNachricht postfachNachrichtWithNoReferencedNachricht = PostfachNachrichtTestFactory.createBuilder()
				.referencedNachricht(null).build();
		private final PostfachNachricht postfachNachrichtWithNonMatchingReferencedNachricht = PostfachNachrichtTestFactory.createBuilder()
				.referencedNachricht("quatsch").build();
		private final PostfachNachricht matchingPostfachNachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void mock() {
			when(vorgangService.findVorgangIds(any(), any())).thenReturn(Stream.of(MessageTestFactory.VORGANG_ID));
			when(attachedItemRemoteService.findByVorgangId(any())).thenReturn(Stream.of(postfachNachrichtWithNoReferencedNachricht,
					matchingPostfachNachricht, postfachNachrichtWithNonMatchingReferencedNachricht));
		}

		@Test
		void shouldCallVorgangRemoteService() {
			doFindAnswers().toList();

			verify(vorgangService).findVorgangIds(serviceKontoType, PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
		}

		@Test
		void shouldCallAttachedItemRemoteService() {
			doFindAnswers().toList();

			verify(attachedItemRemoteService).findByVorgangId(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldReturnMatchingPostfachNachricht() {
			var answers = doFindAnswers();

			assertThat(answers).containsExactly(matchingPostfachNachricht);
		}

		private Stream<PostfachNachricht> doFindAnswers() {
			return service.findAnswers(serviceKontoType, PostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE,
					GrpcPostfachMailTestFactory.REFERENCED_NACHRICHT_ID);
		}
	}
}
